#!/usr/bin/env python3

import argparse
import subprocess
import sys
import os
import json
import glob
import re
import time

NVME_CLI = 'nvme'
NVME_SUBSYS_CONFIGFS = '/sys/class/nvme-subsystem/'
SPDK_RPC = 'spdk_rpc.py'


def run_command(command):
    '''
    Run shell command
    @param command: command to run
    @return: result
    '''
    cmd_list = command.split()
    result = subprocess.run(cmd_list, shell=False,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE)
    return result


def load_module(module):
    '''
    Load kernel module
    @param module: module name to load
    '''
    cmd = 'modprobe {}'.format(module)
    result = run_command(cmd)

    if result.returncode:
        sys.exit(result.stderr.decode('utf-8'))


def load_nvme_transport_module(transport):
    '''
    Load kernel driver based on transport
    type
    @param transport: transport type
    '''
    if transport == 'rdma':
        load_module('nvme-rdma')
    elif transport == 'tcp':
        load_module('nvme-tcp')


def set_nvme_io_policy(policy, subsys_nqn):
    '''
    Set nvme I/O policy
    @param policy: policy to set
    @param subsys_nqn: subsystem nqn name
    '''
    subsystems = glob.glob(NVME_SUBSYS_CONFIGFS + '*/')
    for subsystem in subsystems:
        with open(subsystem + 'subsysnqn') as s:
            if s.read().splitlines()[0] == subsys_nqn:
                io_policy_path = '{}iopolicy'.format(subsystem)
                if os.path.exists(io_policy_path):
                    with open(io_policy_path, 'w') as p:
                        p.write(policy)
                break


def convert_path_to_dict(path, parser):
    '''
    Convert connection path to dictionary
    @param path: path to convert
    @return: dict with parsed path
    '''
    result = dict()

    try:
        for item in (path.split('/')):
            k, v = item.split('=')
            result[k] = v
    except ValueError:
        parser.error('wrong PATH format! ({})'.format(path))

    return result


def parse_paths(parser, paths):
    '''
    Parse paths to remote target
    @param parser: parser object
    @param paths: list of paths
    @return: list with formatted paths
    '''
    formatted_paths = []

    path_keys = ['transport', 'adrfam',
                 'traddr', 'trsvcid']

    for path in paths:
        formatted_path = convert_path_to_dict(path, parser)
        for path_key in path_keys:
            if path_key not in formatted_path:
                parser.error('wrong PATH format! {} is missing)'.format(path_key))
        if formatted_path['transport'] not in ['rdma', 'tcp']:
            parser.error('wrong PATH format! (transport should be one of rdma/tcp)')
        formatted_paths.append(formatted_path)

    return formatted_paths


def nvme_list():
    '''
    List nvme devices exposed on host
    @return: devices list
    '''
    cmd = NVME_CLI + ' list -o json'
    result = run_command(cmd)

    if result.returncode:
        sys.exit(result.stderr.decode('utf-8'))

    if result.stdout:
        result_json = json.loads(result.stdout)

        devices = [
                {
                    'type': 'kernel',
                    'dev': dev["DevicePath"],
                    'block_size': dev["SectorSize"]
                }
                for dev in result_json["Devices"]
        ]

        return devices

    return []


def nvme_list_formatted(nvme_list):
    '''
    Format nvme list
    @return: list of formatted
    nvme devices
    '''
    return ','.join(
        [
            ':'.join(
                    ['{}={}'.format(key, value) for key, value in dev.items()]
                    ) for dev in nvme_list
            ]
            )


def connect_nvme(transport, nqn, traddr, trsvcid, hostnqn):
    '''
    Connect nvme device from remote target
    @param transport: transport type
    @param nqn: nqn name
    @param traddr: transport address
    @param trsvcid: transport service
    @param hostnqn: user-defined hostnqn
    '''
    cmd = NVME_CLI + ' connect --transport={} \
                        --traddr={} \
                        --trsvcid={} \
                        --nqn={} \
                        --hostnqn={}'.format(transport,
                                             traddr,
                                             trsvcid,
                                             nqn,
                                             hostnqn)

    result = run_command(cmd)

    if result.returncode:
        sys.exit(result.stderr.decode('utf-8'))


def discover_nvme(transport, traddr, trsvcid, hostnqn=None):
    '''
    Run nvme discover command
    @param transport: transport type
    @param traddr: transport address
    @param trsvcid: transport service id
    @param hostnqn: user-defined hostnqn
    @return: raw output command result
    '''
    cmd = NVME_CLI + ' discover --transport={} \
                         --traddr={} \
                         --trsvcid={} \
                         --raw=/dev/stdout'.format(transport,
                                                   traddr,
                                                   trsvcid)

    if hostnqn:
        cmd += ' --hostnqn={}'.format(hostnqn)

    result = run_command(cmd)
    if result.returncode:
        sys.exit(result.stderr.decode('utf-8'))

    output_tail = "Discovery log is saved to /dev/stdout"
    print(result.stdout[:-(len(output_tail) + 1)].hex())


def disconnect_nvme_subsys(nqn):
    '''
    Disconnect from NVMeoF subsystem
    @param nqn: nqn name
    '''
    cmd = NVME_CLI + ' disconnect --nqn {}'.format(nqn)

    result = run_command(cmd)

    if result.returncode:
        sys.exit(result.stderr.decode('utf-8'))

    print(result.stdout.decode('utf-8'))


def nvme_list_subsys(dev=None):
    '''
    List nvme subsystems
    @param dev: device
    @return: json formatted
    nvme subsystems
    '''
    cmd = NVME_CLI + ' list-subsys -o json'

    if dev:
        cmd += ' {}'.format(dev)

    result = run_command(cmd)

    if result.returncode:
        sys.exit(result.stderr.decode('utf-8'))

    if result.stdout:
        ret = json.loads(result.stdout.decode('utf-8'))
        if dev:
            # nvme-cli <= 1.8.1 bug: Paths added as new subsystem
            # https://github.com/linux-nvme/nvme-cli/pull/504
            if len(ret['Subsystems']) > 1 and 'Paths' in ret['Subsystems'][1]:
                paths = ret['Subsystems'][1]['Paths']
                del ret['Subsystems'][1]
                ret['Subsystems'][0]['Paths'] = paths
        return ret

    return {}


def nvme_get_bdev_info_formatted(bdev):
    '''
    Get nvme device information in
    formatted output
    @param bdev: device
    @return: formatted list with
    device's paths
    '''

    paths = nvme_list_subsys(bdev)['Subsystems'][0]['Paths']
    result = ','.join(['transport={} adrfam=ipv4 {}'.format(path['Transport'], path['Address']).replace(' ', '/') for path in paths])

    return result


def spdk_nvme_get_controllers():
    '''
    SPDK NVMe controllers list
    @return: json object with
    spdk nvme controllers
    '''
    cmd = SPDK_RPC + ' bdev_nvme_get_controllers'

    result = run_command(cmd)

    if result.returncode:
        sys.exit(result.stdout.decode('utf-8'))

    return json.loads(result.stdout.decode('utf-8'))


def spdk_nvme_attach_controller(name, trtype, traddr,
                                adrfam, trsvcid, subnqn,
                                hostnqn=None):
    '''
    SPDK attach nvme controller
    @param name: name of NVMe controller (prefix)
    @param trtype: NVMe-oF target trtype
    @param adrfam: NVMe-oF target adrfam
    @param trsvcid: NVMe-oF target trsvcid
    @param subnqn: NVMe-oF target subnqn
    @param hostnqn: NVMe-oF host subnqn
    '''
    cmd = SPDK_RPC + ' bdev_nvme_attach_controller --name {} \
                      --trtype {} --traddr {} --adrfam {} \
                      --trsvcid {} --subnqn {}'.format(name,
                                                       trtype,
                                                       traddr,
                                                       adrfam,
                                                       trsvcid,
                                                       subnqn)
    if hostnqn:
        cmd += ' --hostnqn {}'.format(hostnqn)

    result = run_command(cmd)

    if result.returncode:
        sys.exit(result.stdout.decode('utf-8').replace('\n', ' '))

    spdk_nvme_bdevs = result.stdout.decode('utf-8').split()

    spdk_nvme_bdevs_detail = [
        spdk_get_bdevs(bdev)[0] for bdev in spdk_nvme_bdevs
        ]

    print(','.join(
        ['dev={}/block_size={}'.format(bdev['name'],
                                       bdev['block_size']) for bdev in spdk_nvme_bdevs_detail]
        ))


def spdk_nvme_get_controller_name():
    '''
    Calculate next available NVMe
    controller name to be created
    @return: NVMe controller name
    string
    '''
    spdk_nvme_controllers = [
        nvme_controller['name'] for nvme_controller in spdk_nvme_get_controllers()
        ]

    nvme_pattern = re.compile('^Nvme(\d+)$')
    nvme_indexes = [
        int(nvme_pattern.search(dev).group(1)) for dev in spdk_nvme_controllers if nvme_pattern.match(dev)
        ]

    if len(nvme_indexes) > 0:
        nvme_index = sorted(nvme_indexes)[-1] + 1
        return 'Nvme{}'.format(nvme_index)

    return 'Nvme0'


def spdk_bdev_nvme_detach_controller(ctrl):
    '''
    Detach an NVMe controller and
    delete any associated bdevs
    @param ctrl: Name of the controller
    '''
    cmd = SPDK_RPC + ' bdev_nvme_detach_controller {}'.format(ctrl)

    result = run_command(cmd)

    if result.returncode:
        sys.exit(result.stdout.decode('utf-8'))


def spdk_bdev_uring_create(filename, name, block_size):
    '''
    Create a bdev with io_uring backend
    @param filename: Path to device or file (ex: /dev/nvme0n1)
    @param name: bdev name
    @param block_size: Block size for this bdev
    '''
    cmd = SPDK_RPC + ' bdev_uring_create {} {} {}'.format(filename,
                                                          name,
                                                          block_size)
    result = run_command(cmd)

    return result


def spdk_bdev_uring_delete(name):
    '''
    Delete a uring bdev
    @param name: uring bdev name
    '''
    cmd = SPDK_RPC + ' bdev_uring_delete {}'.format(name)

    result = run_command(cmd)

    if result.returncode:
        sys.exit(result.stdout.decode('utf-8'))


def spdk_bdev_aio_create(filename, name, block_size):
    '''
    Add a bdev with aio backend
    @param filename: Path to device or file (ex: /dev/sda)
    @param name: Block device name
    @param block_size: Block size for this bdev
    '''
    cmd = SPDK_RPC + ' bdev_aio_create {} {} {}'.format(filename,
                                                        name,
                                                        block_size)
    result = run_command(cmd)

    return result


def spdk_bdev_aio_delete(name):
    '''
    Delete an aio disk
    @param name: aio bdev name
    '''
    cmd = SPDK_RPC + ' bdev_aio_delete {}'.format(name)

    result = run_command(cmd)

    if result.returncode:
        sys.exit(result.stdout.decode('utf-8'))


def spdk_get_bdevs(name=None):
    '''
    Show SPDK block devices
    @return: json object with
    block devices
    '''
    cmd = SPDK_RPC + ' bdev_get_bdevs'
    if name:
        cmd += ' --name {}'.format(name)

    result = run_command(cmd)

    if result.returncode:
        sys.exit(result.stdout.decode('utf-8'))

    return json.loads(result.stdout.decode('utf-8'))


def spdk_get_bdev_info_formatted(bdev):
    '''
    Show SPDK block device path information
    in formatted output
    @return: formatted string with device path
    '''
    bdev_info = spdk_get_bdevs(bdev)

    if 'aio' in bdev_info[0]['driver_specific']:
        bdev_filename = bdev_info[0]['driver_specific']['aio']['filename']
        return nvme_get_bdev_info_formatted(bdev_filename)
    elif 'uring' in bdev_info[0]['driver_specific']:
        bdev_filename = bdev_info[0]['driver_specific']['uring']['filename']
        return nvme_get_bdev_info_formatted(bdev_filename)
    elif 'nvme' in bdev_info[0]['driver_specific']:
        trtype = bdev_info[0]['driver_specific']['nvme']['trid']['trtype'].lower()
        adrfam = bdev_info[0]['driver_specific']['nvme']['trid']['adrfam'].lower()
        traddr = bdev_info[0]['driver_specific']['nvme']['trid']['traddr']
        trsvcid = bdev_info[0]['driver_specific']['nvme']['trid']['trsvcid']

        return 'transport={}/adrfam={}/traddr={}/trsvcid={}'.format(trtype,
                                                                    adrfam,
                                                                    traddr,
                                                                    trsvcid)
    else:
        sys.exit('not supported driver type for device {}'.format(bdev))


def main():
    parser = argparse.ArgumentParser(description='Configure multipath device')
    parser.add_argument('--op',
                        type=str,
                        choices=['connect', 'disconnect', 'discover', 'get_bdev_info'],
                        required=True,
                        help='action (required)')
    parser.add_argument('--protocol',
                        type=str,
                        choices=['nvme'],
                        required=True,
                        help='protocol to use (required)')
    parser.add_argument('--policy',
                        type=str,
                        choices=['numa', 'round-robin'],
                        default='numa',
                        help='i/o policy (default: numa)')
    parser.add_argument('--qn',
                        type=str,
                        help='qn name')
    parser.add_argument('--hostqn',
                        type=str,
                        help='user-defined hostqn')
    parser.add_argument('--path',
                        type=str,
                        action='append',
                        dest='paths',
                        metavar='PATH',
                        help='''
                        connection path
                        (ex. --path=transport=rdma/adrfam=ipv4/traddr=1.1.1.1/trsvcid=4420
                        --path=transport=rdma/adrfam=ipv4/traddr=2.2.2.2/trsvcid=4420)
                        '''
                        )
    parser.add_argument('--dev',
                        type=str,
                        help='namespace name (ex. Nvme0n1)')
    parser.add_argument('--type',
                        type=str,
                        choices=['kernel', 'spdk'],
                        help='namespace type')

    args = parser.parse_args()

    if os.geteuid() != 0:
        sys.exit('root privileges are required to run this script!')

    if args.protocol == 'nvme' and args.op == 'connect':
        if (
            args.paths is None or args.qn is None
           ):
            parser.error('--op=connect and --protocol=nvme require '
                         '--path --qn options.'
                         )

        parsed_paths = parse_paths(parser, args.paths)
        if len(parsed_paths) > 1:
            # Use kernel nvme for multipath
            for path in parsed_paths:
                load_nvme_transport_module(path['transport'])
                connect_nvme(path['transport'], args.qn,
                             path['traddr'], path['trsvcid'],
                             args.hostqn)
                set_nvme_io_policy(args.policy, args.qn)

            # Wait when system will initialize all nvme devices
            # that belong to the current subnqn
            n = 0
            while True:
                nvme_devices = [dev for dev in nvme_list() if nvme_list_subsys(dev['dev'])['Subsystems'][0]['NQN'] == args.qn]
                if len(nvme_devices) == n:
                    break
                n = len(nvme_devices)
                time.sleep(1)

            spdk_nvme_bdevs = []
            for nvme_dev in nvme_devices:
                fn, name, bs = nvme_dev['dev'], nvme_dev['dev'].split('/')[-1], nvme_dev['block_size']
                # Try to create SPDK URING device on top of kernel nvme
                result = spdk_bdev_uring_create(fn, name, bs)
                if result.returncode and 'Method not found' in result.stdout.decode('utf-8'):
                    # If URING isn't supported, try AIO
                    result = spdk_bdev_aio_create(fn, name, bs)

                if result.returncode:
                    sys.exit(result.stdout.decode('utf-8'))

                spdk_nvme_bdevs.append('dev={}/block_size={}'.format(name, bs))
            print(','.join(spdk_nvme_bdevs))
        else:
            path = parsed_paths[0]
            nvme_controller_name = spdk_nvme_get_controller_name()
            spdk_nvme_attach_controller(
                nvme_controller_name, path['transport'],
                path['traddr'], path['adrfam'],
                path['trsvcid'], args.qn,
                args.hostqn
                )

    elif args.protocol == 'nvme' and args.op == 'disconnect':
        if args.qn is None:
            parser.error('--protocol=nvme and --op=disconnect require '
                         '--qn option'
                         )
        nvme_deleted = False
        spdk_nvme_controllers = spdk_nvme_get_controllers()
        for nvme_ctrl in spdk_nvme_controllers:
            if args.qn == nvme_ctrl['trid']['subnqn']:
                spdk_bdev_nvme_detach_controller(nvme_ctrl['name'])
                nvme_deleted = True
                break
        if not nvme_deleted:
            nvme_dev_list = nvme_list()
            if nvme_dev_list:
                for nvme in nvme_dev_list:
                    nvme_sybsys = nvme_list_subsys(nvme['dev'])
                    if nvme_sybsys and args.qn == nvme_sybsys['Subsystems'][0]['NQN']:
                        nvme_dev_prefix = nvme_sybsys['Subsystems'][0]['Paths'][0]['Name']
                        nvme_pattern = re.compile('^/dev/{}.*'.format(nvme_dev_prefix))
                        for bdev_type in ['uring', 'aio']:
                            for spdk_dev in spdk_get_bdevs():
                                if bdev_type in spdk_dev['driver_specific']:
                                    bdev_filename = spdk_dev['driver_specific'][bdev_type]['filename']
                                    if nvme_pattern.match(bdev_filename):
                                        if bdev_type == 'uring':
                                            spdk_bdev_uring_delete(spdk_dev['name'])
                                        elif bdev_type == 'aio':
                                            spdk_bdev_aio_delete(spdk_dev['name'])

                disconnect_nvme_subsys(args.qn)

    elif args.protocol == 'nvme' and args.op == 'discover':
        if (
            args.paths is None
           ):
            parser.error('--op=discover and --protocol=nvme require '
                         '--path option.'
                         )
        if len(args.paths) != 1:
            parser.error('only one path is allowable with --op=discover')

        parsed_paths = parse_paths(parser, args.paths)
        path = parsed_paths[0]
        load_nvme_transport_module(path['transport'])

        discover_nvme(path['transport'], path['traddr'],
                      path['trsvcid'], args.hostqn
                      )

    elif args.protocol == 'nvme' and args.op == 'get_bdev_info':
        if (
            args.dev is None or args.type is None
           ):
            parser.error(
                            '--op=get_bdev_info and --protocol=nvme require '
                            '--dev --type options.'
                        )
        if args.type == 'spdk':
            print(spdk_get_bdev_info_formatted(args.dev))

        elif args.type == 'kernel':
            print(nvme_get_bdev_info_formatted(args.dev))


if __name__ == '__main__':
    main()
